/*
 * Written by Milan Straka as a part of Factoring polynomials
 * over finite fields (fapof) project.
 * Copyright (C) 2006  Milan Straka
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

#ifndef PRIMAL_FINITE_FIELD_H_BY_MILAN_STRAKA_2006
#define PRIMAL_FINITE_FIELD_H_BY_MILAN_STRAKA_2006

#include "../common.h"

template<int id>
class ffprimal {
 public:
  ffprimal() {}

  template<int id_> friend ffprimal<id_> operator+(ffprimal<id_> a,ffprimal<id_> b);
  template<int id_> friend ffprimal<id_> operator-(ffprimal<id_> a,ffprimal<id_> b);
  template<int id_> friend ffprimal<id_> operator*(ffprimal<id_> a,ffprimal<id_> b);
  template<int id_> friend int operator==(ffprimal<id_> a,ffprimal<id_> b);
  template<int id_> friend int operator!=(ffprimal<id_> a,ffprimal<id_> b);
  template<int id_> friend int operator<(ffprimal<id_> a,ffprimal<id_> b);
  template<int id_> friend int operator<=(ffprimal<id_> a,ffprimal<id_> b);
  template<int id_> friend int operator>(ffprimal<id_> a,ffprimal<id_> b);
  template<int id_> friend int operator>=(ffprimal<id_> a,ffprimal<id_> b);
  static ffprimal<id> invert(ffprimal<id> a);
  unsigned bitTest(unsigned b) const { return v&(1<<b); }
  unsigned bitHighest() const { unsigned b=0,tmp=v>>1; while (tmp) tmp>>=1,b++; return b;}

  bool read(char *str) { return (parse_unsigned(str,&v)) ? (v%=p,1) : (v=0,0); }
  void write(void) { printf("%u",v); }
  void random(void) { v=::rand()%ffprimal<id>::p; } // well, not really uniform, but we dont care
  unsigned toUnsigned() const { return v; }
  static ffprimal<id> fromUnsigned(unsigned d) { return ffprimal(d); }

  static bool initFromStr(char *str);
  static bool initRandom(unsigned bits);
  template<int id_> static bool initFFT(unsigned maxdeg, ffprimal<id_> p, ffprimal<id> &w);
  static const ffprimal<id> fchar(void) { return ffprimal(p); }
  static const int           fext(void) { return           1; }
  static const ffprimal<id> fsize(void) { return ffprimal(p); }
  static const ffprimal<id> fzero(void) { return ffprimal(0); }
  static const ffprimal<id>  fone(void) { return ffprimal(1); }
  //FFT stuff
  typedef ffprimal<-id> T_FFT;
  template<int id_> void set(ffprimal<id_> a) { v=a.toUnsigned() % p; }
  template<int id_> void set_without_mod(ffprimal<id_> a) { v=a.toUnsigned(); }
 private:
  ffprimal(unsigned v) { this->v=v; }
  unsigned v;

  static unsigned p;
};


#define CONSTRUCT_FFPRIMAL(name,id)		\
  typedef ffprimal<id> name;			\
  template<> unsigned ffprimal<id>::p=0;	\
  template<> unsigned ffprimal<-id>::p=0;

template<int id> 
inline ffprimal<id> operator+(ffprimal<id> a,ffprimal<id> b) {
  unsigned res=a.v+b.v; 
  if (res >= ffprimal<id>::p) res-=ffprimal<id>::p;
  return res;
}

template<int id> 
inline ffprimal<id> operator-(ffprimal<id> a,ffprimal<id> b) {
  unsigned res=ffprimal<id>::p + a.v - b.v;
  if (res >= ffprimal<id>::p) res-=ffprimal<id>::p;
  return res;
}

template<int id> 
inline ffprimal<id> operator*(ffprimal<id> a,ffprimal<id> b) {
  unsigned res;
  asm ( "mov %1,%%eax\n\tmull %2\n\tdivl %3"
	: "=&d"(res)
	: "g"(a.v), "g"(b.v), "g"(ffprimal<id>::p)
        : "cc", "eax"
      );

  return res;
}

template<int id>
inline ffprimal<id> ffprimal<id>::invert(ffprimal<id> a) {
  unsigned b=ffprimal<id>::p-2;
  if (b==0) return 1;

  ffprimal<id> res = (b&1) ? a : 1;
  while (b>>=1,b) {
    a=a*a;
    if (b&1) res=res*a;
  }
  return res;
}

template<int id>
bool ffprimal<id>::initFromStr(char *str) {
  unsigned pv;
  if (!parse_unsigned(str,&pv)) {ERROR("Error parsing size of finite field '%s'!\n",str);return false;}

  int ok=1;
  if (pv>=(1U<<31)) {ERROR("Finite field with %u elements is too large!\n",pv);return false;}

  if (pv<2 || (pv>2 && (pv%2)==0)) ok=0;
  for (unsigned q=3;q*q<=pv;q+=2)
    if ((pv%q)==0) ok=0;

  if (!ok) {ERROR("Unsupported nonprimal finite field size %u!\n",pv);return false;}
  ffprimal<id>::p=pv;
  return true;
}

template<int id>
bool ffprimal<id>::initRandom(unsigned bits) {
  if (bits<=1) {ffprimal<id>::p=2; /* special case*/ return true;}
  if (bits >= 32) {ERROR("Finite field of %u bits is too large!\n",bits);return false;}
  p=1U<<bits;

  int ok;
  do {
    p--;
    ok=1;
    if ((p%2)==0) ok=0;
    for (unsigned q=3;q*q<=p;q+=2)
      if ((p%q)==0) ok=0;
  } while (!ok);

  return true;
}

template<int id> template<int id_> 
bool ffprimal<id>::initFFT(unsigned maxdeg, ffprimal<id_> p, ffprimal<id> &w) {
  if (maxdeg & (maxdeg-1)) 
    {ERROR("Primal FFT works only when degree is power of two, not with %u!\n",maxdeg);return false;}
  //check that maxdeg <= 2^27
  if (maxdeg > (1U<<27)) 
    {ERROR("Primal FFT works only with polynomials of degree <= %d!\n",1<<27);return false;}

  //check that maxdeg * p * p <= 2013265921
  unsigned div=2013265921 / maxdeg,psize=p.fsize().toUnsigned();
  //sqrt(2013265921) = 44869
  if (psize > 44869 || psize*psize > div)
    {ERROR("Primal FFT works only when (field characteristic)^2 * maxdeg <= %d!\n",2013265921);return false;}

  //ok, set p and "compute" w
  ffprimal<id>::p=2013265921; //15*2^27 + 1, has generator 31
  //2^27-th primitive root of unity is 440564289 = (31^15) [31 is generator]
  unsigned mocnina=(1U<<27) / maxdeg; //mocnina is 2^?
  w=440564289;
  while (mocnina>>=1) w=w*w;
  return true;
}


template<int id> inline int operator==(ffprimal<id> a,ffprimal<id> b) { return a.v == b.v; }
template<int id> inline int operator!=(ffprimal<id> a,ffprimal<id> b) { return a.v != b.v; }
template<int id> inline int operator< (ffprimal<id> a,ffprimal<id> b) { return a.v <  b.v; }
template<int id> inline int operator<=(ffprimal<id> a,ffprimal<id> b) { return a.v <= b.v; }
template<int id> inline int operator> (ffprimal<id> a,ffprimal<id> b) { return a.v >  b.v; }
template<int id> inline int operator>=(ffprimal<id> a,ffprimal<id> b) { return a.v >= b.v; }


#endif
