/*
 * Written by Milan Straka as a part of Factoring polynomials
 * over finite fields (fapof) project.
 * Copyright (C) 2006  Milan Straka
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

#ifdef POLYNOMIAL_H_INCLUDING_PROTOTYPES
private:
  struct ComposeWrapper {
  public:
    ComposeWrapper(const polynom& pa,const polynom& pb,const inverse& pi) : a(pa),b(pb),i(pi) {}
    const polynom &a,&b;const inverse &i;
  };
  void compose(polynom_coefs *to, polynom_coefs *gori, polynom_coefs *hori, 
	       polynom_coefs *finv, polynom_coefs *f);
public:
  static ComposeWrapper compose(const polynom& a,const polynom& b,const inverse& i)
    { return ComposeWrapper(a,b,i); }
  void operator=(const ComposeWrapper& c);
  void pthroot(unsigned v);
#endif

#ifdef POLYNOMIAL_H_INCLUDING_DATA_DECL
  unsigned compose_m;
  polynom_coefs *compose_g;
  polynom_coefs **compose_h;
  T* compose_tmp;
#endif

#ifdef POLYNOMIAL_H_INCLUDING_DATA_CREATE
  data.compose_m=usqrt(deg);
  data.compose_h=new polynom_coefs*[data.compose_m+1];
  for (unsigned i=0;i<=data.compose_m;i++) data.compose_h[i]=Alloc::getmem(deg);
  data.compose_g=Alloc::getmem((2*deg > deg+2*data.compose_m) ? 2*deg : deg+2*data.compose_m);
  data.compose_tmp=new T[data.compose_m];
#endif

#ifdef POLYNOMIAL_H_INCLUDING_DATA_DESTROY
  if (data.compose_h) {
    for (unsigned i=0;i<=data.compose_m;i++) Alloc::freemem(data.compose_h[i]);
    delete[] data.compose_h;
  }
  if (data.compose_g) Alloc::freemem(data.compose_g);
  if (data.compose_tmp) delete[] data.compose_tmp; data.compose_tmp=NULL;
#endif

#ifdef POLYNOMIAL_H_INCLUDING_FUNCTIONS
template<class T, int id>
void polynomial<T,id>:: compose(polynom_coefs *to, polynom_coefs *gori, polynom_coefs *hori, 
				polynom_coefs *finv, polynom_coefs *f) {
  //we want g(h) mod f
  polynom_coefs *g=data.compose_g,**h=data.compose_h;
  T* tmp=data.compose_tmp;

  if (f->deg==0) {to->deg=0;to->c[0]=T::fzero();return;}
  unsigned n=f->deg,m=usqrt(f->deg);

  //compute all h's
  h[0]->deg=0;h[0]->c[0]=T::fone();
  for (unsigned i=1;i<=m;i++) {mul(g,h[i-1],hori); mod(h[i],g,finv,f); }

  //lets get g working... just copy gori & extend it, so that g[m-1] is whole allocated
  CLONE_COEFS(g,gori);
  while (g->deg < n+2*m) g->c[++g->deg]=T::fzero();
  
  //matrix multiplication directly in G.H... compute one column and replace it
  unsigned col=n; //why not for (unsignedcol=n;col>=0;col--) ... col is UNSIGNED
  while (1) {
    for (int j=m-1;j>=0;j--) tmp[j]=T::fzero();
    
    for (int i=m-1;i>=0;i--) if (h[i]->deg >= col) { //this is int (m=sqrt(n), no problem)
      T htmp=h[i]->c[col];
      T *gtmp=g->c+(m-1)*m+i;
      for (int j=m-1;j>=0;j--) {
	tmp[j]=tmp[j] + *gtmp * htmp;
	gtmp-=m;
      }
    }

    for (int j=m-1;j>=0;j--) h[j]->c[col]=tmp[j];
    if (col==0) break; col--;
  }

  //set correct degrees of h[i] & degrify
  for (int i=m-1;i>=0;i--) {h[i]->deg=n; DEGRIFY_COEFS(h[i]); }

  //now just evaluate \sum h[i] * h[m]^i % f
  CLONE_COEFS(p,h[0]);
  if (m==1) {degrify(); return;}
  CLONE_COEFS(h[0],h[m]); //h[0]=1*h[m] -- small speedup
  mul(g,h[0],h[1]);
  mod(h[1],g,finv,f); //now h[i]=h_old[i] * h[m]^i
  add(p,p,h[1]);
  for (unsigned i=2;i<m;i++) {
    mul(g,h[0],h[m]);
    mod(h[0],g,finv,f); //now h[0]=h[m]^i
    mul(g,h[0],h[i]);
    mod(h[i],g,finv,f); //now h[i]=h_old[i] * h[m]^i
    add(p,p,h[i]);
  }

  //as usual
  DEGRIFY_COEFS(to);
}

template<class T, int id>
void polynomial<T,id>::operator=(const ComposeWrapper& c) {
  compose(p,c.a.p,c.b.p,c.i.p,c.i.b);
}

//this is f^{1/v}, kind of "inverse" to f(x^k).
template<class T, int id>
inline void polynomial<T,id>::pthroot(unsigned v) { 
  for (unsigned d=1,s=p->deg/v;d<=s;d++) p->c[d]=p->c[d*v]; 
  p->deg/=v; 
}

#endif
