/*
 * Written by Milan Straka as a part of Factoring polynomials
 * over finite fields (fapof) project.
 * Copyright (C) 2006  Milan Straka
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

#ifdef POLYNOMIAL_H_INCLUDING_PROTOTYPES
private:
  struct InverseWrapper {
  public:
    InverseWrapper(const polynom& pa,unsigned d) : a(pa),max_deg2div(d) {}
    const polynom &a;
    unsigned max_deg2div;
  };
public:
  class inverse {
  public:
    inverse() : p(NULL),b(NULL) {}
    ~inverse() { if (p) Alloc::freemem(p); p=NULL; if (b) Alloc::freemem(b); b=NULL; }
    void operator=(const InverseWrapper &c);
    int deg() const { return b->deg; }
    friend class polynomial<T,id>;
  private:
    void invert(polynom_coefs *orig, unsigned max_deg2div);
    polynom_coefs *p;
    polynom_coefs *b;
  };
  static InverseWrapper invert(const polynom& a,unsigned max_deg2div) { return InverseWrapper(a,max_deg2div); }

private:
  struct DivWrapper {
  public:
    DivWrapper(const polynom& pa,const inverse& pb) : a(pa),b(pb) {}
    const polynom &a;const inverse &b;
  };
  struct ModWrapper {
  public:
    ModWrapper(const polynom& pa,const inverse& pb) : a(pa),b(pb) {}
    const polynom &a;const inverse &b;
  };
  static void div(polynom_coefs *to, polynom_coefs *a, polynom_coefs *inv, polynom_coefs *b);
  static void mod(polynom_coefs *to, polynom_coefs *a, polynom_coefs *inv, polynom_coefs *b);

public:
  DivWrapper operator/(const inverse& b) const { return DivWrapper(*this,b); }
  ModWrapper operator%(const inverse& b) const { return ModWrapper(*this,b); }
  void operator=(const DivWrapper& c);
  void operator=(const ModWrapper& c);
  void operator=(const inverse& b) { CLONE_COEFS(p, b.p); }

#endif

#ifdef POLYNOMIAL_H_INCLUDING_DATA_DECL
  polynom_coefs *divtmp,*divbigtmp;
#endif

#ifdef POLYNOMIAL_H_INCLUDING_DATA_CREATE
  data.divtmp=Alloc::getmem(deg);
  data.divbigtmp=Alloc::getmem(3*deg);
#endif

#ifdef POLYNOMIAL_H_INCLUDING_DATA_DESTROY
  if (data.divtmp) Alloc::freemem(data.divtmp); data.divtmp=NULL;
  if (data.divbigtmp) Alloc::freemem(data.divbigtmp); data.divbigtmp=NULL;
#endif

#ifdef POLYNOMIAL_H_INCLUDING_FUNCTIONS

template<class T, int id>
inline void polynomial<T,id>::inverse::invert(polynom_coefs *orig, unsigned max_deg2div) {
  polynom_coefs *f=data.divtmp,*tmp2=data.multmp,*tmp3=data.divbigtmp;
  unsigned orig_deg=orig->deg;
  unsigned deg=max_deg2div-orig_deg;
  if (b) Alloc::freemem(b);  b=Alloc::getmem(orig);
  if (p) Alloc::freemem(p);  if (max_deg2div < orig_deg) {p=NULL; return;}
  p=Alloc::getmem(deg);

  //1) reverse input polynomial to f
  f->deg=orig_deg;
  for (unsigned i=0;i<=orig_deg;i++) f->c[orig_deg-i]=orig->c[i];

  //lets go inverting...
  p->deg=0;p->c[0]=T::invert(f->c[0]);
  for (unsigned i=1;(1U<<(i-1)) < deg+1;i++) {
    f->deg=((1U<<i) < orig_deg) ? (1U<<i) : orig_deg;
    mul(tmp2,p,p);
    if (tmp2->deg > (1U<<i)) tmp2->deg=1U<<i;
    if (tmp2->deg > deg) tmp2->deg=deg;
    mul(tmp3,tmp2,f);
    if (tmp3->deg > (1U<<i)) tmp3->deg=1U<<i;
    if (tmp3->deg > deg) tmp3->deg=deg;
    for (unsigned i=p->deg;i>0;i--) p->c[i]=p->c[i]+p->c[i]; p->c[0]=p->c[0]+p->c[0];
    sub(p,p,tmp3);
  }
}

template<class T, int id>
inline void polynomial<T,id>::inverse::operator=(const InverseWrapper& c) {
  invert(c.a.p, c.max_deg2div);
}


template<class T, int id>
void polynomial<T,id>::div(polynom_coefs *to, polynom_coefs *a, polynom_coefs *inv, polynom_coefs *b) {
  polynom_coefs *tmp=data.divtmp, *tmp3=data.divbigtmp;
  if (a->deg < b->deg) {to->deg=0;to->c[0]=T::fzero();return;}
  
  unsigned wanted_deg=a->deg-b->deg;
  unsigned orig_inv_deg=inv->deg;
  if (inv->deg > wanted_deg) inv->deg=wanted_deg;

  //now reverse orig
  tmp->deg=wanted_deg;
  {
    unsigned i,j;
    for (i=a->deg-wanted_deg,j=wanted_deg;j>0;j--,i++) tmp->c[j]=a->c[i]; 
    tmp->c[j]=a->c[i];
  }

  //create pseudoquotient
  mul(tmp3,tmp,inv);
  for (unsigned i=tmp3->deg+1;i<=wanted_deg;i++) tmp3->c[i]=T::fzero();
  tmp3->deg=wanted_deg;

  //revert the pseudoquotient
  to->deg=wanted_deg;
  for (unsigned i=wanted_deg;i>0;i--) to->c[wanted_deg-i]=tmp3->c[i]; to->c[wanted_deg]=tmp3->c[0];
  DEGRIFY_COEFS(to);

  inv->deg=orig_inv_deg;
}

template<class T, int id>
inline void polynomial<T,id>::operator=(const DivWrapper& c) {
  div(p,c.a.p,c.b.p,c.b.b);
}


template<class T, int id>
void polynomial<T,id>::mod(polynom_coefs *to, polynom_coefs *a, polynom_coefs *inv, polynom_coefs *b) {
  if (a->deg < b->deg) {CLONE_COEFS(to,a);return;}
  div(data.divtmp,a,inv,b);
  mul(data.divbigtmp,data.divtmp,b);
  sub(to,a,data.divbigtmp);
  DEGRIFY_COEFS(to);
}

template<class T, int id>
inline void polynomial<T,id>::operator=(const ModWrapper& c) {
  mod(p,c.a.p,c.b.p,c.b.b);
}

#endif
