/*
 * Written by Milan Straka as a part of Factoring polynomials
 * over finite fields (fapof) project.
 * Copyright (C) 2006  Milan Straka
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

#ifdef POLYNOMIAL_H_INCLUDING_PROTOTYPES
private:
  typedef typename T::T_FFT T_FFT;
  static unsigned highest_bit(unsigned a) { unsigned res=0; while(a>>=1) res++; return res; }
  static unsigned nearest_2power(unsigned a) { unsigned add=(a&(a-1))!=0; return 1<<(highest_bit(a)+add); }
  static unsigned reverse_bits(unsigned a, unsigned how_much) { //is automatically inline
    a=((a&(0xAAAAAAAA)) >> 1) | ((a&(0x55555555)) << 1);
    a=((a&(0xCCCCCCCC)) >> 2) | ((a&(0x33333333)) << 2);
    a=((a&(0xF0F0F0F0)) >> 4) | ((a&(0x0F0F0F0F)) << 4);
    a=((a&(0xFF00FF00)) >> 8) | ((a&(0x00FF00FF)) << 8);
    a=((a&(0xFFFF0000)) >>16) | ((a&(0x0000FFFF)) <<16);
    return a>>(32-how_much);
  }

  static void fft(unsigned N, T_FFT *what, T_FFT *ws);
#endif

#ifdef POLYNOMIAL_H_INCLUDING_DATA_DECL
  T_FFT* fft_ws, *fft_iws, *fft_a, *fft_b;
  unsigned fft_ws_bits; //means that fft_ws[1] is 2^fft_ws_bits primitive root
                        //and there are [0..2^(fft_ws_bits-1)-1] omegas in _ws and _iws
                                              //we need only half of them
#endif

#ifdef POLYNOMIAL_H_INCLUDING_DATA_CREATE  
  if (deg) {
    unsigned maxdeg=nearest_2power(2*deg+1)>>1; //we need only half of omegas
    data.fft_ws=new T_FFT[maxdeg];   data.fft_iws=new T_FFT[maxdeg]; 
    data.fft_a=new T_FFT[maxdeg<<1]; data.fft_b=new T_FFT[maxdeg<<1]; //but full a and b
    data.fft_ws_bits=highest_bit(maxdeg)+1;                           //and full info in ws_bits
    data.fft_ws[0]=T_FFT::fone();    data.fft_iws[0]=T_FFT::fone();
    if (!T_FFT::initFFT(maxdeg<<1,T::fsize(),data.fft_ws[1])) return false; //and full in initFFT
    data.fft_iws[1]=T_FFT::invert(data.fft_ws[1]);
    for (unsigned i=2;i<maxdeg;i++) {                                 //and only half omegas
      data.fft_ws [i]=data.fft_ws [1] * data.fft_ws [i-1];
      data.fft_iws[i]=data.fft_iws[1] * data.fft_iws[i-1];
    }
    LOGMORE("FFT multiplication for deg %u over GF(%u) initialised.\n", 
	    maxdeg<<1,T_FFT::fsize().toUnsigned());
  } else data.fft_a=data.fft_b=data.fft_ws=data.fft_iws=NULL;
#endif

#ifdef POLYNOMIAL_H_INCLUDING_DATA_DESTROY
  if (data.fft_ws) delete[] data.fft_ws; data.fft_ws=NULL;
  if (data.fft_iws) delete[] data.fft_iws; data.fft_iws=NULL;
  if (data.fft_a) delete[] data.fft_a; data.fft_a=NULL;
  if (data.fft_b) delete[] data.fft_b; data.fft_b=NULL;
#endif

#ifdef POLYNOMIAL_H_INCLUDING_FUNCTIONS
template<class T, int id>
inline void polynomial<T,id>::operator=(const MulWrapper& c) {
  mul(p,c.a.p,c.b.p);
}

template<class T, int id>
void polynomial<T,id>::fft(unsigned N, T_FFT *what, T_FFT *ws) {
  int halfbs=1,bs=2,nb=(1<<(N-1)); //blocksize, numblocks
  unsigned wshift=data.fft_ws_bits-N;
  T_FFT tmp;
  while (N--) {
    for (int i=(nb-1)*bs;i>=0;i-=bs) {
      for (int j=halfbs-1,wj=(j*nb)<<wshift;j>=0;j--,wj-=nb<<wshift) {
	tmp=what[i+halfbs+j]*ws[wj];
	what[i+halfbs+j]=what[i+j]-tmp;
	what[i+j]=what[i+j]+tmp;
      }
    }
    halfbs=bs, bs<<=1, nb>>=1;
  }
}

template<class T, int id>
void polynomial<T,id>::mul(polynom_coefs *to,polynom_coefs *a,polynom_coefs *b) {
  DEGRIFY_COEFS(a);
  DEGRIFY_COEFS(b);
  if (a->deg == 0 && a->c[0]==T::fzero()) {to->deg=0;to->c[0]=T::fzero();return;}
  if (b->deg == 0 && b->c[0]==T::fzero()) {to->deg=0;to->c[0]=T::fzero();return;}
#ifndef BRUTALMUL
  if (a->deg < 4 || b->deg < 4 || a->deg * b->deg <= 25) {
    if (to==a || to==b) { polynom_coefs *tmp=data.multmp; mul_force(tmp,a,b); CLONE_COEFS(to,tmp); }
    else mul_force(to,a,b); return;}
#else
  if (a->deg < 12 && b->deg < 12 && (a->deg < 6 || b->deg < 6)) {
    if (to==a || to==b) { polynom_coefs *tmp=data.multmp; mul_force(tmp,a,b); CLONE_COEFS(to,tmp); }
    else mul_force(to,a,b); return;}
#endif

  unsigned deg=nearest_2power(a->deg + b->deg + 1), degb=highest_bit(deg),i;
  T_FFT *fft_a=data.fft_a, *fft_b=data.fft_b;
  //reverse a to pa & do fft on fa
  for (i=0;i<=a->deg;i++) fft_a[reverse_bits(i,degb)].set_without_mod(a->c[i]);
  for (;i<deg;i++) fft_a[reverse_bits(i,degb)].set_without_mod(T::fzero());
  fft(degb,fft_a,data.fft_ws);

  //reverse b to pb & do fft on fb
  for (i=0;i<=b->deg;i++) fft_b[reverse_bits(i,degb)].set_without_mod(b->c[i]);
  for (;i<deg;i++) fft_b[reverse_bits(i,degb)].set_without_mod(T::fzero());
  fft(degb,fft_b,data.fft_ws);

  //mul pa & pb, reverse pa to pb, defft pb
  for (i=0;i<deg;i++) fft_a[i]=fft_a[i]*fft_b[i];
  for (i=0;i<deg;i++) fft_b[reverse_bits(i,degb)]=fft_a[i];
  fft(degb,fft_b,data.fft_iws);

  //mul each pb by deg^-1 and copy pb to to
  //the set operation does value%fchar, we need it now
  to->deg=a->deg+b->deg;
  T_FFT deginv=T_FFT::invert(T_FFT::fromUnsigned(deg));
  for (i=to->deg;i>0;i--) fft_b[i]=fft_b[i]*deginv; fft_b[0]=fft_b[0]*deginv;  
  for (i=to->deg;i>0;i--) to->c[i].set(fft_b[i]);to->c[0].set(fft_b[0]);
  DEGRIFY_COEFS(to);
}

#endif
