/*
 * Written by Milan Straka as a part of Factoring polynomials
 * over finite fields (fapof) project.
 * Copyright (C) 2006  Milan Straka
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

#ifndef POLYNOMIAL_H_BY_MILAN_STRAKA_2006
#define POLYNOMIAL_H_BY_MILAN_STRAKA_2006

#include <cctype>
#include <cstring>

#include "../common.h"
#include "common.h"
#include "allocator.h"

template<class T, int id>
class polynomial {
 public:
  typedef polynomial<T,id> polynom;
  typedef polynomial_coefs<T> polynom_coefs;

  polynomial(unsigned deg);
  polynomial(const polynom& a);
  polynomial(unsigned deg,const polynom& a);
  const polynom& operator=(const polynom& a);
  ~polynomial();
  polynomial() {STOP("Default constructor for polynomial called!\n");}

  #define POLYNOMIAL_H_INCLUDING_PROTOTYPES
  #include "clone.h"
  #include "compare.h"
  #include "degrify.h"
  #include "monify.h"
  #include "derivate.h"
  #include "random.h"
  #include "addsub.h"
  #include "mul.h"
  #include "divmod.h"
  #include "exp.h"
  #include "gcd.h"
  #include "compose.h"
  #undef POLYNOMIAL_H_INCLUDING_PROTOTYPES

 public:
  bool read(char *str,unsigned max_deg);
  void write(void);
  unsigned deg() const { return p->deg; }

  static bool init(unsigned deg);
  static void destroy();
  static polynom pzero(void) { return *zero; }
  static polynom  pone(void) { return *one; }
  static polynom   pex(void) { return *ex; }
 private:
  typedef alloc<T,id> Alloc;

  polynomial_coefs<T> *p;
  static polynom *zero,*one,*ex;

  struct data_type {
  #define POLYNOMIAL_H_INCLUDING_DATA_DECL
  #include "mul.h"
  #include "divmod.h"
  #include "exp.h"
  #include "gcd.h"
  #include "compose.h"
  #undef POLYNOMIAL_H_INCLUDING_DATA_DECL
  };
  static data_type data;
};


#define CONSTRUCT_POLYNOMIAL(name,T,id)			\
  typedef polynomial<T,id> name;			\
  template<> name* name::zero=NULL;			\
  template<> name* name::one=NULL;			\
  template<> name* name::ex=NULL;			\
  template<> name::data_type name::data=data_type();	\
  CONSTRUCT_ALLOCATOR(T,id)

template<class T, int id>
bool polynomial<T,id>::init(unsigned deg) {
  Alloc::init(3*deg); // tripple, because in polynomial inversions we really need 3*deg

  zero=new polynom(0);
  zero->p->deg=0;zero->p->c[0]=T::fzero();
  one=new polynom(0);
  one->p->deg=0;one->p->c[0]=T::fone();  
  ex=new polynom(1);
  ex->p->deg=1;ex->p->c[0]=T::fzero();ex->p->c[1]=T::fone();

  #define POLYNOMIAL_H_INCLUDING_DATA_CREATE
  #include "mul.h"
  #include "divmod.h"
  #include "exp.h"
  #include "gcd.h"
  #include "compose.h"
  #undef POLYNOMIAL_H_INCLUDING_DATA_CREATE

  return true;
}

template<class T, int id>
void polynomial<T,id>::destroy() {
  #define POLYNOMIAL_H_INCLUDING_DATA_DESTROY
  #include "mul.h"
  #include "divmod.h"
  #include "exp.h"
  #include "gcd.h"
  #include "compose.h"
  #undef POLYNOMIAL_H_INCLUDING_DATA_DESTROY

  if (one) delete one; one=NULL;
  if (zero) delete zero; zero=NULL;
  if (ex) delete ex; ex=NULL;

  Alloc::destroy();
}

// MEMORY FUNCTIONS
template<class T, int id>
inline polynomial<T,id>::polynomial(unsigned deg) {
  p=Alloc::getmem(deg);
  p->deg=0;
  p->c[0]=T::fzero();
}

template<class T, int id>
inline polynomial<T,id>::polynomial(const polynomial<T,id>& a) {
  p=Alloc::getmem(a.p);
}

template<class T, int id>
inline polynomial<T,id>::polynomial(unsigned deg, const polynomial<T,id>& a) {
  p=Alloc::getmem(deg);
  CLONE_COEFS(p,a.p);
}

template<class T, int id>
inline const polynomial<T,id>& polynomial<T,id>::operator=(const polynomial<T,id>& a) {
  if (p!=a.p) {
    Alloc::freemem(p);
    p=Alloc::getmem(a.p);
  }
  return *this;
}

template<class T, int id>
inline polynomial<T,id>::~polynomial() {
  Alloc::freemem(p);
}

// INPUT/OUTPUT

template<class T, int id>
inline bool polynomial<T,id>::read(char *str,unsigned max_deg) {
  for (unsigned i=0;i<=max_deg;i++) p->c[i]=T::fzero();
  p->deg=0;

  while (*str) {
    //+- at the beginning
    while (isspace(*str)) str++;
    int plus=1;
    if (*str=='+') str++;
    else if (*str=='-') plus=0,str++;
    while (isspace(*str)) str++;

    //find next monomial
    char *dalsi=strpbrk(str,"+-"),orig='+';
    if (dalsi) orig=*dalsi,*dalsi='\0';

    //try to find 'x'
    unsigned d=0;
    T tmp;
    char *ex=strpbrk(str,"xX");
    if (ex) {
      char *tmpex=ex;
      *ex++='\0';
      d=1;
      while (isspace(*ex)) ex++;
      if (*ex=='^') if (!parse_unsigned(ex+1,&d)) {ERROR("Cannot parse exponent 'x%s'!\n",ex); return false;}
      if (strlen(str)) {
	if (!tmp.read(str)) {ERROR("Cannot parse coefficient '%s'!\n",str); return false;}
      } else tmp=T::fone();
      *tmpex='x';
    } else {
      if (!tmp.read(str)) {ERROR("Cannot parse coefficient '%s'!\n",str); return false;}
    }

    //hooray, lets use it
    if (d>max_deg) {ERROR("Degree %d too large, expected %d maximum!\n",d,max_deg); return false;}
    p->c[d]=(plus) ? tmp : T::fzero() - tmp;
    if (tmp!=T::fzero() && d>p->deg) p->deg=d;

    //and continue
    if (dalsi) *dalsi=orig,str=dalsi;
    else str="";
  }

  return true;
}

template<class T, int id>
inline void polynomial<T,id>::write(void) {
  int written=0;
  if (!p->deg) p->c[0].write();
  else for (int i=p->deg;i>=0;i--) if (p->c[i]!=T::fzero()) {
    if (written) putchar('+');written=1;
    if (p->c[i]!=T::fone() || i==0) p->c[i].write();
    if (i>0) putchar('x');
    if (i>1) printf("^%d",i);
  }
}

#define POLYNOMIAL_H_INCLUDING_FUNCTIONS
#include "clone.h"
#include "compare.h"
#include "degrify.h"
#include "monify.h"
#include "derivate.h"
#include "random.h"
#include "addsub.h"
#include "mul.h"
#include "divmod.h"
#include "exp.h"
#include "gcd.h"
#include "compose.h"
#undef POLYNOMIAL_H_INCLUDING_FUNCTIONS

#endif
